#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <stdio.h>

#include "response.h"
#include "scan.h"

void tcp_response(const struct tcphdr *tcphdr, const struct scan *data)
{
	if (data->type == SCAN_UDP) {
		dprintf(2, "ft_nmap: error: received a TCP response for an UDP "
		           "scan\n");
		return;
	}
	if (data->type == SCAN_SYN) {
		if (tcphdr->ack == 1 && tcphdr->syn == 1)
			data->response->state = OPEN;
		else if (tcphdr->ack == 1 && tcphdr->rst == 1)
			data->response->state = CLOSE;
	} else if (data->type == SCAN_ACK && tcphdr->rst == 1)
		data->response->state = UNFILTERED;
	else if (data->type == SCAN_NULL && tcphdr->rst == 1)
		data->response->state = CLOSE;
	else if (data->type == SCAN_FIN && tcphdr->rst == 1)
		data->response->state = CLOSE;
	else if (data->type == SCAN_XMAS && tcphdr->rst == 1)
		data->response->state = CLOSE;
}

void udp_response(const struct udphdr *udphdr, const struct scan *data)
{
	(void)udphdr;
	if (data->type != SCAN_UDP) {
		dprintf(2, "ft_nmap: error: received an UDP response for a TCP "
		           "scan\n");
		return;
	}
	data->response->state = OPEN;
}

void icmp_response(const struct icmphdr *icmphdr, const struct scan *data)
{
	if (data->type == SCAN_SYN && icmphdr->type == 3)
		data->response->state = FILTERED;
	else if (data->type == SCAN_ACK && icmphdr->type == 3)
		data->response->state = FILTERED;
	else if (data->type == SCAN_NULL && icmphdr->type == 3)
		data->response->state = FILTERED;
	else if (data->type == SCAN_FIN && icmphdr->type == 3)
		data->response->state = FILTERED;
	else if (data->type == SCAN_XMAS && icmphdr->type == 3)
		data->response->state = FILTERED;
	else if (data->type == SCAN_UDP && icmphdr->type == 3 &&
	         icmphdr->code == 3)
		data->response->state = CLOSE;
	else if (data->type == SCAN_UDP && icmphdr->type == 3 &&
	         icmphdr->code != 3)
		data->response->state = FILTERED;
}
